-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

package.path = "lua/vehicle/?.lua;lua/vehicle/jbeam/?.lua;lua/common/?.lua;lua/?.lua;?.lua"
require('compatibility')

log = function(...) Lua:log(...) end
print = function(...) Lua:log('A', "print", tostring(...)) end

require('utils')
require('mathlib')

math.randomseed(os.time())

-- improve stacktraces
local STP = require "StackTracePlus"
debug.traceback = STP.stacktrace

perf = require("perf")
settings = require('simplesettings')

vehicleMode = 0
lastDt = 1/20

local engineShiftUpRPM = 0

local function nop() end
log_jbeam = nop -- intentionally global

local loaded = false

-- step functions
function physicsStep(dts)
    drivetrain.updateWheels(dts)
    sound_enhance.update(dts) --StinchinStein
    sensors.update(dts)
    thrusters.update()
    hydros.update(dts)
end

-- This is called in the global scope, so it is safe to do things that contact things outside the vehicle
function graphicsStep(dt)
    beamstate.updateGFX()
    bullettime.update(dt)
end

-- This is called in the local scope, so it is NOT safe to do things that contact things outside the vehicle
-- It is called right after graphicsStepf
function graphicsStepLocal(dt)
    if not loaded then return end
    lastDt = dt
    sensors.updateGFX(dt) -- must be before input and ai
    mapmgr.sendTracking() -- must be before ai.updateGFX
    ai.updateGFX(dt) -- must be before input
    input.update(dt) -- must be as early as possible
    electrics.update(dt)
    extensions.hook('updateGFX', dt)
    drivetrain.updateEngine(dt)
    sounds.update(dt)
    hydros.updateGFX(dt)
    thrusters.updateGFX()
    if vehicleMode == 1 then
        recovery.updateGFX(dt)
        perf.update()
        streams.updateGFX()
        gui.updateGFX(dt)
        guihooks.updateGFX()

        -- void playRPMLeds(const float currentRPM, const float rpmFirstLedTurnsOn, const float rpmRedLine)
        --print('playRPMLeds| currentRPM=' .. tostring(drivetrain.rpm) .. ' , rpmFirstLedTurnsOn=' .. tostring(v.data.engine.shiftUpRPM * 0.8) .. ' , rpmRedLine =' .. tostring(v.data.engine.shiftUpRPM))
        --obj:playRPMLeds(drivetrain.rpm, engineShiftUpRPM * 0.8, engineShiftUpRPM)
    end
    drivetrain.resetWheels(dt)
    mapmgr.updateGFX()
    scenario.updateGFX()
    props.update()
    fire.updateGFX(dt)
end

-- debug rendering
function debugDraw(x, y, z)
    local focusPos = float3(0,0,0)
    if x and y and z then focusPos = float3(x,y,z) end
    bdebug.debugDraw(focusPos)
    ai.debugDraw(focusPos)
end

function enableCallbacks()
    -- this enables the most important frame callbacks
    -- in case of an error, they still point to nop and thus do not spam the console
    -- TODO: FIXME
end

function initSystems()
    material.init()
    drivetrain.init()
    sound_enhance.init()
    sensors.reset()
    beamstate.init()
    thrusters.init()
    hydros.init()
    input.init()
    sounds.init()
    props.init()
    electrics.init()
    recovery.init()
    bdebug.init()
    scenario.init()
    sensors.init()
    fire.init()

    gui.message("", 0, "^damage\\.") -- clear damage messages on vehicle restart
    extensions.hook('onInit')
end

function init(path, partConfigData)

    --perf.enable(1)

    if not obj then log('W', "default.init", "Error getting main object: unable to spawn") ; return end
    log('D', "default.init", "spawning vehicle "..tostring(path))
    if partConfigData and string.len(partConfigData) > 0 then
      log('D', "default.init", "partconfig: "..tostring(partConfigData))
    end

    -- we change the lookup path here, so it prefers the vehicle lua
    package.path = path .. "/lua/?.lua;".. package.path
    extensions = require("extensions")
    extensions.loadModules(path.."/lua")
    extensions.addModulePath("lua/vehicle/extensions/")
    
    drivetrain = require("drivetrain")
    sound_enhance = require("sound_enhance")
    
    sounds     = require("sounds")
    bdebug     = require("bdebug")
    input      = require("input")
    props      = require("props")

    particlefilter = require("particlefilter")
    particles  = require("particles")
    material   = require("material")
    v          = require("jbeam_main")
    electrics  = require("electrics")
    json       = require("json")
    beamstate  = require("beamstate")
    sensors    = require("sensors")
    bullettime = require("bullettime")
    thrusters  = require("thrusters")
    hydros     = require("hydros")
    gui        = require("gui") -- do not change its name, the GUI callback will break otherwise
    partmgmt   = require("partmgmt") -- do not change its name, the GUI callback will break otherwise
    streams    = require("guistreams")
    guihooks   = require("guihooks")
    ai         = require("ai")
    recovery   = require("recovery")
    mapmgr     = require("mapmgr")
    scenario   = require("scenario")
    fire       = require("fire")

    local hp = HighPerfTimer()

    -- care about the config before pushing to the physics
    if type(partConfigData) == 'string' and string.len(partConfigData) > 0 then
        local firstChar = string.sub(partConfigData,1,1)
        if firstChar == '{' or firstChar == '[' then
            log('D', "default.init", "using partconfig data "..tostring(partConfigData)..' - path:'..tostring(path) )
            partmgmt.setConfig(unserialize(partConfigData), false)
        else
            log('D', "default.init", "using partconfig filename "..tostring(partConfigData)..' - path:'..tostring(path) )
            partmgmt.load(partConfigData, false)
        end
    end

    -- enable/disable Jbeam logging depending on the devmode
    if settings and settings.getValue('devMode') == true then
      log_jbeam = log
    end

    -- load jbeam file
    if not v.loadDirectories({'vehicles/common/', path}) then -- important: the real vehicle dir is always last, all additional paths come first
        return
    end

    -- you can change the data in here before it gets submitted to the physics

    -- submit to physics
    v.pushToPhysics(obj)

    if v.data == nil then
        v.data = {}
    end

    initSystems()

    -- temporary tire mark setting
    obj.slipTireMarkThreshold = 15

    -- if we got to here, lets enable the callbacks and hope for the best
    enableCallbacks()

    if v.data.engine ~= nil then
        engineShiftUpRPM = v.data.engine.shiftUpRPM or 0
    end

    --perf.disable()
    --perf.saveDataToCSV('vehicleloading.csv')

    -- update gravity
    obj:queueGameEngineLua( 'be:getObjectByID('..tostring(obj:getID())..'):queueLuaCommand(\"obj:setGravity( \"..environment.getGravity()..\" )\")' )

    log('D', "default.init", "init done - vehicle loading took "..hp:stop()..' ms')
    
    loaded = true
end

-- various callbacks
function beamBroken(id, energy)
    beamstate.beamBroken(id, energy)
    --bullettime.slowMotion(0.05, 20) -- 20% of realtime
end

function beamDeformed(id, ratio)
    beamstate.beamDeformed(id, ratio)
end

-- called when everything was cleared
function vehicleDestroy()
    --log('D', "default.vehicleDestroy", "vehicleDestroy()")
    -- when the vehicle gets unloaded, remove all sounds
    sounds.destroy()
    material.destroy()
end

-- called when the user pressed I
function vehicleResetted(retainDebug)
    --log('D', "default.vehicleResetted", "vehicleResetted()")
    electrics.reset()
    sound_enhance.init()
    drivetrain.reset()
    props.reset()
    sensors.reset()
    if not retainDebug then
        bdebug.reset()
    end
    bullettime.reset()
    beamstate.reset()
    thrusters.reset()
    input.reset()
    hydros.reset()
    material.reset()
    guihooks.reset()
    extensions.hook('onReset', retainDebug)
    ai.reset()
    partmgmt.vehicleResetted()
    mapmgr.reset()
    scenario.reset()
    fire.reset()

    gui.message("", 0, "^damage\\.") -- clear damage messages on vehicle restart
    obj:queueGameEngineLua("extensions.hook('onVehicleResetted'," .. obj:getID() .. ")")
end

function nodeCollision(p)
    drivetrain.updateWheelSlip(p)
    fire.nodeCollision(p)
    particlefilter.nodeCollision(p)
    --sounds.nodeCollision(p)
end

function activated(mode)
    vehicleMode = mode
    if vehicleMode == 1 then mapmgr.enableTracking() end
    if vehicleMode == 0 then mapmgr.disableTracking() end
    --[[
    if mode == 1 then
        log('D', "default.activated", "yay, we got active")
    else
        log('D', "default.activated", "deactivated :(")
    end
    ]]--

    bdebug.activated(mode)
    ai.stateChanged()
    extensions.hook('activated', mode)
    guihooks.trigger("VehicleFocusChanged", {id = obj:getID(), mode = mode})
end

function instabilityDetected()
    -- enable break debug, so we see what beams broke
    local msg = "instability detected for vehicle " .. tostring(v.vehicleDirectory) .. ": resetting the vehicle and disabling the physics. "
    log('E', "default.instabilityDetected", msg)

    ui_message(msg, 10, 'instability')

    BeamEngine:setEnabled(false)
    obj:requestReset(RESET_PHYSICS)

   --if settings and settings.getValue('disableDynamicCollision') == true then
   --  Settings.dynamicCollision = false
   --end
end

function exportPersistentData()
    local s = serializeModules()
    --log('D', "default.exportPersistentData", s)
    obj:setPersistentData(s)
end

function importPersistentData(s)
    --log('D', "default.importPersistentData", s)
    -- deserialize extensions first, so the extensions are loaded before they are trying to get deserialized
    deserializeModules(unserialize(s), 'extensions')
    deserializeModules(unserialize(s))
end

function guiUpdate(d)
    --log('D', "default.guiUpdate", "guiUpdate()")
    gui.gotUpdate(d)
end

-- used by the GE camera code
function requestCameraConfig()
  if not cameraData then cameraData = {} end
  obj:queueGameEngineLua("extensions.hook('onVehicleCameraConfigChanged'," .. obj:getID() .. "," .. serialize(cameraData) .. ")")
end
